"""Plan that comprises:
* Disable feedback
* Set undulator energy to the requested amount
* Adjust DCM and mirrors for the new energy
* reenable feedback
"""

import bluesky.preprocessors as bpp
import pydantic
from bluesky import plan_stubs as bps
from dodal.devices.attenuator.attenuator import BinaryFilterAttenuator
from dodal.devices.dcm import DCM
from dodal.devices.focusing_mirror import FocusingMirrorWithStripes, MirrorVoltages
from dodal.devices.undulator import Undulator
from dodal.devices.undulator_dcm import UndulatorDCM
from dodal.devices.xbpm_feedback import XBPMFeedback

from mx_bluesky.common.parameters.constants import PlanNameConstants
from mx_bluesky.common.preprocessors.preprocessors import (
    transmission_and_xbpm_feedback_for_collection_wrapper,
)
from mx_bluesky.hyperion.device_setup_plans import dcm_pitch_roll_mirror_adjuster

DESIRED_TRANSMISSION_FRACTION = 0.1

UNDULATOR_GROUP = "UNDULATOR_GROUP"


@pydantic.dataclasses.dataclass(config={"arbitrary_types_allowed": True})
class SetEnergyComposite:
    vfm: FocusingMirrorWithStripes
    mirror_voltages: MirrorVoltages
    dcm: DCM
    undulator_dcm: UndulatorDCM
    xbpm_feedback: XBPMFeedback
    attenuator: BinaryFilterAttenuator


# Remove composite after https://github.com/DiamondLightSource/dodal/issues/1092
@pydantic.dataclasses.dataclass(config={"arbitrary_types_allowed": True})
class XBPMWrapperComposite:
    undulator: Undulator
    xbpm_feedback: XBPMFeedback
    attenuator: BinaryFilterAttenuator
    dcm: DCM


@bpp.set_run_key_decorator(PlanNameConstants.SET_ENERGY)
@bpp.run_decorator()
def _set_energy_plan(
    energy_kev,
    composite: SetEnergyComposite,
):
    yield from bps.abs_set(composite.undulator_dcm, energy_kev, group=UNDULATOR_GROUP)
    yield from dcm_pitch_roll_mirror_adjuster.adjust_dcm_pitch_roll_vfm_from_lut(
        composite.undulator_dcm,
        composite.vfm,
        composite.mirror_voltages,
        energy_kev,
    )
    yield from bps.wait(group=UNDULATOR_GROUP)


def set_energy_plan(
    energy_ev: float | None,
    composite: SetEnergyComposite,
):
    # Remove conversion after https://github.com/DiamondLightSource/dodal/issues/1092
    composite_for_wrapper = XBPMWrapperComposite(
        composite.undulator_dcm.undulator_ref._obj,  # noqa: SLF001
        composite.xbpm_feedback,
        composite.attenuator,
        composite.dcm,
    )

    if energy_ev:
        yield from transmission_and_xbpm_feedback_for_collection_wrapper(
            _set_energy_plan(energy_ev / 1000, composite),
            composite_for_wrapper,
            DESIRED_TRANSMISSION_FRACTION,
        )



"""
Extruder data collection
This version in python3 new Feb2021 by RLO
    - March 21 added logging and Eiger functionality
"""

import re
import sys
import time
from datetime import datetime
from pathlib import Path
from pprint import pformat

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
from bluesky.utils import MsgGenerator
from dodal.common import inject
from dodal.devices.attenuator.attenuator import ReadOnlyAttenuator
from dodal.devices.hutch_shutter import HutchShutter, ShutterDemand
from dodal.devices.i24.aperture import Aperture
from dodal.devices.i24.beam_center import DetectorBeamCenter
from dodal.devices.i24.beamstop import Beamstop
from dodal.devices.i24.dcm import DCM
from dodal.devices.i24.dual_backlight import DualBacklight
from dodal.devices.i24.focus_mirrors import FocusMirrorsMode
from dodal.devices.i24.i24_detector_motion import DetectorMotion
from dodal.devices.zebra.zebra import Zebra

from mx_bluesky.beamlines.i24.serial.dcid import (
    DCID,
    get_pilatus_filename_template_from_device,
    read_beam_info_from_hardware,
)
from mx_bluesky.beamlines.i24.serial.log import (
    SSX_LOGGER,
    _read_visit_directory_from_file,
    log_on_entry,
)
from mx_bluesky.beamlines.i24.serial.parameters import ExtruderParameters
from mx_bluesky.beamlines.i24.serial.parameters.constants import BEAM_CENTER_LUT_FILES
from mx_bluesky.beamlines.i24.serial.setup_beamline import Pilatus, caget, caput, pv
from mx_bluesky.beamlines.i24.serial.setup_beamline import setup_beamline as sup
from mx_bluesky.beamlines.i24.serial.setup_beamline.setup_detector import (
    UnknownDetectorType,
    get_detector_type,
)
from mx_bluesky.beamlines.i24.serial.setup_beamline.setup_zebra_plans import (
    GATE_START,
    arm_zebra,
    disarm_zebra,
    open_fast_shutter,
    reset_zebra_when_collection_done_plan,
    set_shutter_mode,
    setup_zebra_for_extruder_with_pump_probe_plan,
    setup_zebra_for_quickshot_plan,
)
from mx_bluesky.beamlines.i24.serial.write_nexus import call_nexgen

SAFE_DET_Z = 1480


def flush_print(text):
    sys.stdout.write(str(text))
    sys.stdout.flush()


@log_on_entry
def initialise_extruder(
    detector_stage: DetectorMotion = inject("detector_motion"),
) -> MsgGenerator:
    SSX_LOGGER.info("Initialise Parameters for extruder data collection on I24.")

    visit = caget(pv.ioc13_gp1)
    SSX_LOGGER.info(f"Visit defined {visit}")

    # Define detector in use
    det_type = yield from get_detector_type(detector_stage)

    caput(pv.ioc13_gp2, "test")
    caput(pv.ioc13_gp3, "testrun")
    caput(pv.ioc13_gp4, "100")
    caput(pv.ioc13_gp5, "0.01")
    caput(pv.ioc13_gp6, 0)
    caput(pv.ioc13_gp8, 0)  # status PV do not reuse gp8 for something else
    caput(pv.ioc13_gp9, 0)
    caput(pv.ioc13_gp10, 0)
    caput(pv.ioc13_gp15, det_type.name)
    caput(pv.pilat_cbftemplate, 0)
    SSX_LOGGER.info("Initialisation complete.")
    yield from bps.null()


@log_on_entry
def laser_check(
    mode: str,
    zebra: Zebra = inject("zebra"),
    detector_stage: DetectorMotion = inject("detector_motion"),
) -> MsgGenerator:
    """Plan to open the shutter and check the laser beam from the viewer by pressing \
        'Laser On' and 'Laser Off' buttons on the edm.

    The 'Laser on' button sets the correct OUT_TTL pv for the detector in use to \
    SOFT_IN1 and the shutter mode to auto.
    The 'Laser off' button disconnects the OUT_TTL pv set by the previous step and \
    resets the shutter mode to manual.

    WARNING. When using the laser with the extruder, some hardware changes need to be made.
    Because all four of the zebra ttl outputs are in use in this mode, when the \
    detector in use is the Eiger, the Pilatus cable is repurposed to trigger the light \
    source, and viceversa.
    """
    SSX_LOGGER.debug(f"Laser check: {mode}")

    det_type = yield from get_detector_type(detector_stage)

    LASER_TTL = (
        zebra.mapping.outputs.TTL_EIGER
        if isinstance(det_type, Pilatus)
        else zebra.mapping.outputs.TTL_PILATUS
    )

    if mode == "laseron":
        yield from bps.abs_set(
            zebra.output.out_pvs[LASER_TTL], zebra.mapping.sources.SOFT_IN3
        )
        yield from set_shutter_mode(zebra, "auto")

    if mode == "laseroff":
        yield from bps.abs_set(
            zebra.output.out_pvs[LASER_TTL], zebra.mapping.sources.DISCONNECT
        )
        yield from set_shutter_mode(zebra, "manual")


@log_on_entry
def enter_hutch(
    detector_stage: DetectorMotion = inject("detector_motion"),
) -> MsgGenerator:
    """Move the detector stage before entering hutch."""
    yield from bps.mv(detector_stage.z, SAFE_DET_Z)
    SSX_LOGGER.debug("Detector moved.")


@log_on_entry
def read_parameters(detector_stage: DetectorMotion, attenuator: ReadOnlyAttenuator):
    """ Read the parameters from user input and create the parameter model for an \
        extruder collection.

    Args:
        detector_stage (DetectorMotion): The detector stage device.
        attenuator (ReadOnlyAttenuator): A read-only attenuator device to get the \
            transmission value.

    Returns:
        ExtruderParameters: Parameter model for extruder collections

    """
    SSX_LOGGER.info("Creating parameter model from input.")

    det_type = yield from get_detector_type(detector_stage)
    SSX_LOGGER.warning(f"DETECTOR TYPE: {det_type}")
    filename = caget(pv.ioc13_gp3)
    # If file name ends in a digit this causes processing/pilatus pain.
    # Append an underscore
    if det_type.name == "pilatus":
        m = re.search(r"\d+$", filename)
        if m is not None:
            # Note for future reference. Appending underscore causes more hassle and
            # high probability of users accidentally overwriting data. Use a dash
            filename = filename + "-"
            SSX_LOGGER.info(
                f"Requested filename ends in a number. Appended dash: {filename}"
            )

    transmission = yield from bps.rd(attenuator.actual_transmission)

    pump_status = bool(int(caget(pv.ioc13_gp6)))
    pump_exp = float(caget(pv.ioc13_gp9)) if pump_status else 0.0
    pump_delay = float(caget(pv.ioc13_gp10)) if pump_status else 0.0

    params_dict = {
        "visit": _read_visit_directory_from_file().as_posix(),  # noqa
        "directory": caget(pv.ioc13_gp2),
        "filename": filename,
        "exposure_time_s": float(caget(pv.ioc13_gp5)),
        "detector_distance_mm": float(caget(pv.ioc13_gp7)),
        "detector_name": str(det_type),
        "transmission": transmission,
        "num_images": int(caget(pv.ioc13_gp4)),
        "pump_status": pump_status,
        "laser_dwell_s": pump_exp,
        "laser_delay_s": pump_delay,
    }

    SSX_LOGGER.info("Parameters \n")
    SSX_LOGGER.info(pformat(params_dict))
    yield from bps.null()
    return ExtruderParameters(**params_dict)


@log_on_entry
def main_extruder_plan(
    zebra: Zebra,
    aperture: Aperture,
    backlight: DualBacklight,
    beamstop: Beamstop,
    detector_stage: DetectorMotion,
    shutter: HutchShutter,
    dcm: DCM,
    mirrors: FocusMirrorsMode,
    beam_center_device: DetectorBeamCenter,
    parameters: ExtruderParameters,
    dcid: DCID,
    start_time: datetime,
) -> MsgGenerator:
    beam_center_pixels = sup.compute_beam_center_position_from_lut(
        BEAM_CENTER_LUT_FILES[parameters.detector_name],
        parameters.detector_distance_mm,
        parameters.detector_size_constants,
    )
    yield from sup.set_detector_beam_center_plan(
        beam_center_device,
        beam_center_pixels,
    )

    # Setting up the beamline
    SSX_LOGGER.info("Open hutch shutter")
    yield from bps.abs_set(shutter, ShutterDemand.OPEN, wait=True)

    yield from sup.setup_beamline_for_collection_plan(
        aperture, backlight, beamstop, wait=True
    )

    yield from sup.move_detector_stage_to_position_plan(
        detector_stage, parameters.detector_distance_mm
    )

    # For pixel detector
    filepath = parameters.collection_directory.as_posix()
    SSX_LOGGER.debug(f"Filepath {filepath}")
    SSX_LOGGER.debug(f"Filename {parameters.filename}")

    if parameters.detector_name == "pilatus":
        SSX_LOGGER.info("Using pilatus mini cbf")
        caput(pv.pilat_cbftemplate, 0)
        SSX_LOGGER.info(f"Pilatus quickshot setup: filepath {filepath}")
        SSX_LOGGER.info(f"Pilatus quickshot setup: filepath {parameters.filename}")
        SSX_LOGGER.info(
            f"Pilatus quickshot setup: number of images {parameters.num_images}"
        )
        SSX_LOGGER.info(
            f"Pilatus quickshot setup: exposure time {parameters.exposure_time_s}"
        )

        if parameters.pump_status:
            SSX_LOGGER.info("Pump probe extruder data collection")
            SSX_LOGGER.info(f"Pump exposure time {parameters.laser_dwell_s}")
            SSX_LOGGER.info(f"Pump delay time {parameters.laser_delay_s}")
            yield from sup.pilatus(
                "fastchip",
                [
                    filepath,
                    parameters.filename,
                    parameters.num_images,
                    parameters.exposure_time_s,
                ],
            )
            yield from setup_zebra_for_extruder_with_pump_probe_plan(
                zebra,
                parameters.detector_name,
                parameters.exposure_time_s,
                parameters.num_images,
                parameters.laser_dwell_s,
                parameters.laser_delay_s,
                pulse1_delay=0.0,
                wait=True,
            )
        else:
            SSX_LOGGER.info("Static experiment: no photoexcitation")
            yield from sup.pilatus(
                "quickshot",
                [
                    filepath,
                    parameters.filename,
                    parameters.num_images,
                    parameters.exposure_time_s,
                ],
            )
            yield from setup_zebra_for_quickshot_plan(
                zebra, parameters.exposure_time_s, parameters.num_images, wait=True
            )

    elif parameters.detector_name == "eiger":
        SSX_LOGGER.info("Using Eiger detector")

        SSX_LOGGER.debug(f"Creating the directory for the collection in {filepath}.")

        caput(pv.eiger_seqID, int(caget(pv.eiger_seqID)) + 1)
        SSX_LOGGER.info(f"Eiger quickshot setup: filepath {filepath}")
        SSX_LOGGER.info(f"Eiger quickshot setup: filepath {parameters.filename}")
        SSX_LOGGER.info(
            f"Eiger quickshot setup: number of images {parameters.num_images}"
        )
        SSX_LOGGER.info(
            f"Eiger quickshot setup: exposure time {parameters.exposure_time_s}"
        )

        if parameters.pump_status:
            SSX_LOGGER.info("Pump probe extruder data collection")
            SSX_LOGGER.debug(f"Pump exposure time {parameters.laser_dwell_s}")
            SSX_LOGGER.debug(f"Pump delay time {parameters.laser_delay_s}")
            yield from sup.eiger(
                "triggered",
                [
                    filepath,
                    parameters.filename,
                    parameters.num_images,
                    parameters.exposure_time_s,
                ],
            )
            yield from setup_zebra_for_extruder_with_pump_probe_plan(
                zebra,
                parameters.detector_name,
                parameters.exposure_time_s,
                parameters.num_images,
                parameters.laser_dwell_s,
                parameters.laser_delay_s,
                pulse1_delay=0.0,
                wait=True,
            )
        else:
            SSX_LOGGER.info("Static experiment: no photoexcitation")
            yield from sup.eiger(
                "quickshot",
                [
                    filepath,
                    parameters.filename,
                    parameters.num_images,
                    parameters.exposure_time_s,
                ],
            )
            yield from setup_zebra_for_quickshot_plan(
                zebra, parameters.exposure_time_s, parameters.num_images, wait=True
            )
    else:
        err = f"Unknown Detector Type, det_type = {parameters.detector_name}"
        SSX_LOGGER.error(err)
        raise UnknownDetectorType(err)

    beam_settings = yield from read_beam_info_from_hardware(
        dcm, mirrors, beam_center_device, parameters.detector_name
    )

    # Do DCID creation BEFORE arming the detector
    if parameters.detector_name == "eiger":
        filetemplate = f"{parameters.filename}.nxs"
    else:
        filetemplate = yield from get_pilatus_filename_template_from_device()
    dcid.generate_dcid(
        beam_settings=beam_settings,
        image_dir=parameters.collection_directory.as_posix(),
        file_template=filetemplate,
        num_images=parameters.num_images,
        start_time=start_time,
        pump_probe=parameters.pump_status,
    )

    # Collect
    SSX_LOGGER.info("Fast shutter opening")
    yield from open_fast_shutter(zebra)
    if parameters.detector_name == "pilatus":
        SSX_LOGGER.info("Pilatus acquire ON")
        caput(pv.pilat_acquire, 1)
    elif parameters.detector_name == "eiger":
        SSX_LOGGER.info("Triggering Eiger NOW")
        caput(pv.eiger_trigger, 1)

    dcid.notify_start()

    if parameters.detector_name == "eiger":
        SSX_LOGGER.debug("Call nexgen server for nexus writing.")
        beam_x = yield from bps.rd(beam_center_device.beam_x)
        beam_y = yield from bps.rd(beam_center_device.beam_y)
        call_nexgen(
            None,
            parameters,
            beam_settings.wavelength_in_a,
            (beam_x, beam_y),
            start_time,
        )

    timeout_time = time.time() + parameters.num_images * parameters.exposure_time_s + 10

    yield from arm_zebra(zebra)
    yield from bps.sleep(
        GATE_START
    )  # bps.sleep for the same length of gate_start, hard coded to 1
    i = 0
    text_list = ["|", "/", "-", "\\"]
    while True:
        line_of_text = "\r\t\t\t Waiting   " + 30 * (f"{text_list[i % 4]}")
        flush_print(line_of_text)
        yield from bps.sleep(0.5)
        i += 1
        zebra_arm_status = yield from bps.rd(zebra.pc.arm.armed)
        if zebra_arm_status == 0:  # not zebra.pc.is_armed():
            # As soon as zebra is disarmed, exit.
            # Epics updates this PV once the collection is done.
            SSX_LOGGER.info("Zebra disarmed - Collection done.")
            break
        if time.time() >= timeout_time:
            SSX_LOGGER.warning(
                """
                Something went wrong and data collection timed out. Aborting.
            """
            )
            raise TimeoutError("Data collection timed out.")

    SSX_LOGGER.info("Collection completed without errors.")


@log_on_entry
def collection_aborted_plan(
    zebra: Zebra, detector_name: str, dcid: DCID
) -> MsgGenerator:
    """A plan to run in case the collection is aborted before the end."""
    SSX_LOGGER.warning("Data Collection Aborted")
    yield from disarm_zebra(zebra)  # If aborted/timed out zebra still armed
    if detector_name == "pilatus":
        caput(pv.pilat_acquire, 0)
    elif detector_name == "eiger":
        caput(pv.eiger_acquire, 0)
    yield from bps.sleep(0.5)
    end_time = datetime.now()
    dcid.collection_complete(end_time, aborted=True)


@log_on_entry
def tidy_up_at_collection_end_plan(
    zebra: Zebra,
    shutter: HutchShutter,
    parameters: ExtruderParameters,
    dcid: DCID,
) -> MsgGenerator:
    """A plan to tidy up at the end of a collection, successful or aborted.

    Args:
        zebra (Zebra): The Zebra device.
        shutter (HutchShutter): The HutchShutter device.
        parameters (ExtruderParameters): Collection parameters.
    """
    yield from reset_zebra_when_collection_done_plan(zebra)

    # Clean Up
    if parameters.detector_name == "pilatus":
        yield from sup.pilatus("return-to-normal", None)
    elif parameters.detector_name == "eiger":
        yield from sup.eiger("return-to-normal", None)
        SSX_LOGGER.debug(f"{parameters.filename}_{caget(pv.eiger_seqID)}")
    SSX_LOGGER.debug("End of Run")
    SSX_LOGGER.info("Close hutch shutter")
    yield from bps.abs_set(shutter, ShutterDemand.CLOSE, wait=True)

    dcid.notify_end()


@log_on_entry
def collection_complete_plan(
    collection_directory: Path, detector_name: str, dcid: DCID
) -> MsgGenerator:
    if detector_name == "pilatus":
        SSX_LOGGER.info("Pilatus Acquire STOP")
        caput(pv.pilat_acquire, 0)
    elif detector_name == "eiger":
        SSX_LOGGER.info("Eiger Acquire STOP")
        caput(pv.eiger_acquire, 0)
        caput(pv.eiger_ODcapture, "Done")

    yield from bps.sleep(0.5)

    end_time = datetime.now()
    dcid.collection_complete(end_time, aborted=False)
    SSX_LOGGER.info(f"End Time = {end_time.ctime()}")

    yield from bps.null()


def run_extruder_plan(
    zebra: Zebra = inject("zebra"),
    aperture: Aperture = inject("aperture"),
    backlight: DualBacklight = inject("backlight"),
    beamstop: Beamstop = inject("beamstop"),
    detector_stage: DetectorMotion = inject("detector_motion"),
    shutter: HutchShutter = inject("shutter"),
    dcm: DCM = inject("dcm"),
    mirrors: FocusMirrorsMode = inject("focus_mirrors"),
    attenuator: ReadOnlyAttenuator = inject("attenuator"),
) -> MsgGenerator:
    start_time = datetime.now()
    SSX_LOGGER.info(f"Collection start time: {start_time.ctime()}")

    parameters: ExtruderParameters = yield from read_parameters(
        detector_stage, attenuator
    )
    # Create collection directory
    parameters.collection_directory.mkdir(parents=True, exist_ok=True)

    beam_center_device = sup.get_beam_center_device(parameters.detector_name)

    # DCID - not generated yet
    dcid = DCID(emit_errors=False, expt_params=parameters)

    yield from bpp.contingency_wrapper(
        main_extruder_plan(
            zebra=zebra,
            aperture=aperture,
            backlight=backlight,
            beamstop=beamstop,
            detector_stage=detector_stage,
            shutter=shutter,
            dcm=dcm,
            mirrors=mirrors,
            beam_center_device=beam_center_device,
            parameters=parameters,
            dcid=dcid,
            start_time=start_time,
        ),
        except_plan=lambda e: (
            yield from collection_aborted_plan(zebra, parameters.detector_name, dcid)
        ),
        else_plan=lambda: (
            yield from collection_complete_plan(
                parameters.collection_directory, parameters.detector_name, dcid
            )
        ),
        final_plan=lambda: (
            yield from tidy_up_at_collection_end_plan(zebra, shutter, parameters, dcid)
        ),
        auto_raise=False,
    )



from __future__ import annotations

from typing import TYPE_CHECKING

from mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping import (
    get_proposal_and_session_from_visit_string,
)
from mx_bluesky.common.external_interaction.callbacks.common.plan_reactive_callback import (
    PlanReactiveCallback,
)
from mx_bluesky.common.external_interaction.ispyb.exp_eye_store import (
    BLSampleStatus,
    ExpeyeInteraction,
    RobotActionID,
)
from mx_bluesky.common.utils.log import ISPYB_ZOCALO_CALLBACK_LOGGER
from mx_bluesky.hyperion.parameters.constants import CONST

if TYPE_CHECKING:
    from event_model.documents import Event, EventDescriptor, RunStart, RunStop


class RobotLoadISPyBCallback(PlanReactiveCallback):
    def __init__(self) -> None:
        ISPYB_ZOCALO_CALLBACK_LOGGER.debug("Initialising ISPyB Robot Load Callback")
        super().__init__(log=ISPYB_ZOCALO_CALLBACK_LOGGER)
        self._metadata: dict | None = None

        self.run_uid: str | None = None
        self.descriptors: dict[str, EventDescriptor] = {}
        self.action_id: RobotActionID | None = None
        self.expeye = ExpeyeInteraction()

    def activity_gated_start(self, doc: RunStart):
        ISPYB_ZOCALO_CALLBACK_LOGGER.debug(
            "ISPyB robot load callback received start document."
        )
        if doc.get("subplan_name") == CONST.PLAN.ROBOT_LOAD:
            ISPYB_ZOCALO_CALLBACK_LOGGER.debug(
                f"ISPyB robot load callback received: {doc}"
            )
            self.run_uid = doc.get("uid")
            self._metadata = doc.get("metadata")
            assert isinstance(self._metadata, dict)
            proposal, session = get_proposal_and_session_from_visit_string(
                self._metadata["visit"]
            )
            self.action_id = self.expeye.start_load(
                proposal,
                session,
                self._metadata["sample_id"],
                self._metadata["sample_puck"],
                self._metadata["sample_pin"],
            )
        return super().activity_gated_start(doc)

    def activity_gated_descriptor(self, doc: EventDescriptor) -> EventDescriptor | None:
        self.descriptors[doc["uid"]] = doc
        return super().activity_gated_descriptor(doc)

    def activity_gated_event(self, doc: Event) -> Event | None:
        event_descriptor = self.descriptors.get(doc["descriptor"])
        if (
            event_descriptor
            and event_descriptor.get("name") == CONST.DESCRIPTORS.ROBOT_LOAD
        ):
            assert self.action_id is not None, (
                "ISPyB Robot load callback event called unexpectedly"
            )
            barcode = doc["data"]["robot-barcode"]
            oav_snapshot = doc["data"]["oav-snapshot-last_saved_path"]
            webcam_snapshot = doc["data"]["webcam-last_saved_path"]
            # I03 uses webcam/oav snapshots in place of before/after snapshots
            self.expeye.update_barcode_and_snapshots(
                self.action_id, barcode, webcam_snapshot, oav_snapshot
            )

        return super().activity_gated_event(doc)

    def activity_gated_stop(self, doc: RunStop) -> RunStop | None:
        ISPYB_ZOCALO_CALLBACK_LOGGER.debug(
            "ISPyB robot load callback received stop document."
        )
        if doc.get("run_start") == self.run_uid:
            assert self.action_id is not None, (
                "ISPyB Robot load callback stop called unexpectedly"
            )
            exit_status = doc.get("exit_status")
            assert exit_status, "Exit status not available in stop document!"
            assert self._metadata, "Metadata not received before stop document."
            reason = doc.get("reason") or "OK"

            self.expeye.end_load(self.action_id, exit_status, reason)
            self.expeye.update_sample_status(
                self._metadata["sample_id"],
                BLSampleStatus.LOADED
                if exit_status == "success"
                else BLSampleStatus.ERROR_BEAMLINE,
            )
            self.action_id = None
        return super().activity_gated_stop(doc)



import subprocess

cmd = "ssh qqh35939@i03-control"
process = subprocess.Popen(
    cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
)
stdout, stderr = process.communicate()
if process.returncode != 0:
    print(f"Error occurred: {stderr.decode()}")
else:
    print(f"Output: {stdout.decode()}")



from typing import Any

from mx_bluesky.beamlines.i24.serial.fixed_target.ft_utils import ChipType
from mx_bluesky.beamlines.i24.serial.parameters.experiment_parameters import (
    ChipDescription,
)
from mx_bluesky.beamlines.i24.serial.setup_beamline import caget, pv

OXFORD_BLOCKS_PVS = [f"ME14E-MO-IOC-01:GP{i}" for i in range(11, 75)]


class EmptyMapError(Exception):
    pass


def get_chip_format(chip_type: ChipType) -> ChipDescription:
    """Default parameter values."""
    defaults: dict[str, int | float] = {}
    match chip_type:
        case ChipType.Oxford:
            defaults["x_num_steps"] = defaults["y_num_steps"] = 20
            defaults["x_step_size"] = defaults["y_step_size"] = 0.125
            defaults["x_blocks"] = defaults["y_blocks"] = 8
            defaults["b2b_horz"] = defaults["b2b_vert"] = 0.800
        case ChipType.OxfordInner:
            defaults["x_num_steps"] = defaults["y_num_steps"] = 25
            defaults["x_step_size"] = defaults["y_step_size"] = 0.600
            defaults["x_blocks"] = defaults["y_blocks"] = 1
            defaults["b2b_horz"] = defaults["b2b_vert"] = 0.0
        case ChipType.Minichip:
            defaults["x_num_steps"] = defaults["y_num_steps"] = 20
            defaults["x_step_size"] = defaults["y_step_size"] = 0.125
            defaults["x_blocks"] = defaults["y_blocks"] = 1
            defaults["b2b_horz"] = defaults["b2b_vert"] = 0.0
        case ChipType.Custom:
            defaults["x_num_steps"] = int(caget(pv.me14e_gp6))
            defaults["y_num_steps"] = int(caget(pv.me14e_gp7))
            defaults["x_step_size"] = float(caget(pv.me14e_gp8))
            defaults["y_step_size"] = float(caget(pv.me14e_gp99))
            defaults["x_blocks"] = defaults["y_blocks"] = 1
            defaults["b2b_horz"] = defaults["b2b_vert"] = 0.0
        case ChipType.MISP:
            defaults["x_num_steps"] = defaults["y_num_steps"] = 78
            defaults["x_step_size"] = defaults["y_step_size"] = 0.1193
            defaults["x_blocks"] = defaults["y_blocks"] = 1
            defaults["b2b_horz"] = defaults["b2b_vert"] = 0.0
    chip_params: dict[str, Any] = {"chip_type": chip_type, **defaults}
    return ChipDescription(**chip_params)


def get_chip_map() -> list[int]:
    """Return a list of blocks (the 'chip map') to be collected on an Oxford type chip \
        when using lite mapping."""
    chipmap = []
    for n, block_pv in enumerate(OXFORD_BLOCKS_PVS):
        block_val = int(caget(block_pv))
        if block_val == 1:
            chipmap.append(n + 1)
    if len(chipmap) == 0:
        raise EmptyMapError("No blocks selected for Lite map.")
    return chipmap



import os

import pytest

from mx_bluesky.common.utils.utils import convert_angstrom_to_eV
from mx_bluesky.hyperion.parameters.gridscan import HyperionSpecifiedThreeDGridScan
from mx_bluesky.hyperion.parameters.rotation import RotationScan

from ....conftest import (
    default_raw_gridscan_params,
    raw_params_from_file,
)


@pytest.fixture
def test_rotation_params():
    param_dict = raw_params_from_file(
        "tests/test_data/parameter_json_files/good_test_rotation_scan_parameters.json"
    )
    param_dict["storage_directory"] = "tests/test_data"
    param_dict["file_name"] = "TEST_FILENAME"
    param_dict["demand_energy_ev"] = 12700
    param_dict["scan_width_deg"] = 360.0
    params = RotationScan(**param_dict)
    params.x_start_um = 0
    params.y_start_um = 0
    params.z_start_um = 0
    params.exposure_time_s = 0.004
    return params


@pytest.fixture(params=[1050])
def test_fgs_params(request):
    assert request.param % 25 == 0, "Please use a multiple of 25 images"
    params = HyperionSpecifiedThreeDGridScan(**default_raw_gridscan_params())
    params.demand_energy_ev = convert_angstrom_to_eV(1.0)
    params.use_roi_mode = True
    first_scan_img = (request.param // 10) * 6
    second_scan_img = (request.param // 10) * 4
    params.x_steps = 5
    params.y_steps = first_scan_img // 5
    params.z_steps = second_scan_img // 5
    params.storage_directory = (
        os.path.dirname(os.path.realpath(__file__)) + "/test_data"
    )
    params.file_name = "dummy"
    yield params


def default_raw_params(
    json_file="tests/test_data/parameter_json_files/good_test_parameters.json",
):
    return raw_params_from_file(json_file)



import argparse

from pydantic.dataclasses import dataclass

from mx_bluesky._version import version


@dataclass
class HyperionArgs:
    dev_mode: bool = False
    verbose_event_logging: bool = False
    skip_startup_connection: bool = False


def _add_callback_relevant_args(parser: argparse.ArgumentParser) -> None:
    """adds arguments relevant to hyperion-callbacks."""
    parser.add_argument(
        "--dev",
        action="store_true",
        help="Use dev options, such as local graylog instances and S03",
    )


def parse_callback_dev_mode_arg() -> bool:
    """Returns the bool representing the 'dev_mode' argument."""
    parser = argparse.ArgumentParser()
    _add_callback_relevant_args(parser)
    args = parser.parse_args()
    return args.dev


def parse_cli_args() -> HyperionArgs:
    """Parses all arguments relevant to hyperion. Returns an HyperionArgs dataclass with
    the fields: (verbose_event_logging: bool,
                 dev_mode: bool,
                 skip_startup_connection: bool)"""
    parser = argparse.ArgumentParser()
    _add_callback_relevant_args(parser)
    parser.add_argument(
        "--verbose-event-logging",
        action="store_true",
        help="Log all bluesky event documents to graylog",
    )
    parser.add_argument(
        "--skip-startup-connection",
        action="store_true",
        help="Skip connecting to EPICS PVs on startup",
    )
    parser.add_argument(
        "--version",
        help="Print hyperion version string",
        action="version",
        version=version,
    )
    args = parser.parse_args()
    return HyperionArgs(
        verbose_event_logging=args.verbose_event_logging or False,
        dev_mode=args.dev or False,
        skip_startup_connection=args.skip_startup_connection or False,
    )



import logging
import time
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from mx_bluesky.beamlines.i24.serial import log


@pytest.fixture
def dummy_logger():
    logger = logging.getLogger("I24serial")
    yield logger


def _destroy_handlers(logger):
    for handler in logger.handlers:
        handler.close()
    logger.handlers.clear()


@patch("mx_bluesky.beamlines.i24.serial.log.environ")
@patch("mx_bluesky.beamlines.i24.serial.log.Path.mkdir")
def test_logging_file_path(mock_dir, mock_environ):
    mock_environ.get.return_value = None
    log_path = log._get_logging_file_path()
    assert mock_dir.call_count == 1
    assert log_path.as_posix() == "tmp/logs"


@patch("mx_bluesky.beamlines.i24.serial.log._read_visit_directory_from_file")
@patch("mx_bluesky.beamlines.i24.serial.log.environ")
@patch("mx_bluesky.beamlines.i24.serial.log.Path.mkdir")
def test_logging_file_path_on_beamline(mock_dir, mock_environ, mock_visit):
    mock_environ.get.return_value = "i24"
    mock_visit.return_value = Path("/path/to/i24/data")
    log_path = log._get_logging_file_path()
    assert mock_dir.call_count == 1
    assert log_path.as_posix() == "/path/to/i24/data/tmp/serial/logs"


def test_basic_logging_config(dummy_logger):
    assert dummy_logger.hasHandlers() is True
    assert len(dummy_logger.handlers) == 1
    assert dummy_logger.handlers[0].level == logging.INFO


def test_integrate_bluesky_logs():
    mock_dodal_logger = MagicMock()
    with (
        patch("mx_bluesky.beamlines.i24.serial.log.bluesky_logger") as mock_bluesky_log,
        patch(
            "mx_bluesky.beamlines.i24.serial.log.ophyd_async_logger"
        ) as mock_ophyd_log,
    ):
        log._integrate_bluesky_logs(mock_dodal_logger)
        assert mock_bluesky_log.parent == mock_dodal_logger
        assert mock_ophyd_log.parent == mock_dodal_logger


@patch("mx_bluesky.beamlines.i24.serial.log.Path.mkdir")
@patch("mx_bluesky.beamlines.i24.serial.log.do_default_logging_setup")
@patch("mx_bluesky.beamlines.i24.serial.log._integrate_bluesky_logs")
def test_logging_config_with_filehandler(
    mock_integrate_logs, mock_default, mock_dir, dummy_logger
):
    with patch("mx_bluesky.beamlines.i24.serial.log.dodal_logger") as mock_dodal_logger:
        log.config("dummy.log", delayed=True, dev_mode=True)
        assert len(dummy_logger.handlers) == 2
        mock_default.assert_called_once()
        mock_integrate_logs.assert_called_once_with(mock_dodal_logger)
        assert mock_dodal_logger.removeHandler.call_count == 1
        assert mock_dir.call_count == 1
        assert dummy_logger.handlers[1].level == logging.DEBUG
        # Clear FileHandler to avoid other tests failing if it is kept open
        dummy_logger.removeHandler(dummy_logger.handlers[1])
        _destroy_handlers(dummy_logger.parent)


@patch("mx_bluesky.beamlines.i24.serial.log.config")
def test_setup_collection_logs_in_dev_mode(mock_config, RE):
    # Fixed target, dev mode
    fake_filename = time.strftime("i24fixedtarget_%d%B%y.log").lower()
    RE(log.setup_collection_logs("Serial Fixed", True))

    mock_config.assert_called_once_with(fake_filename, dev_mode=True)


@patch("mx_bluesky.beamlines.i24.serial.log.config")
def test_setup_collection_logs(mock_config, RE):
    # Extruder, non dev mode
    fake_filename = time.strftime("i24extruder_%d%B%y.log").lower()
    RE(log.setup_collection_logs("Serial Jet"))

    mock_config.assert_called_once_with(fake_filename, dev_mode=False)


def test_clean_up_log(dummy_logger, RE):
    with patch("mx_bluesky.beamlines.i24.serial.log.dodal_logger") as mock_dodal_logger:
        RE(log.clean_up_log_config_at_end())

        assert len(dummy_logger.handlers) == 0
        assert len(mock_dodal_logger.handlers) == 0



import configparser
from dataclasses import dataclass
from enum import StrEnum

from requests import JSONDecodeError, patch, post
from requests.auth import AuthBase

from mx_bluesky.common.external_interaction.ispyb.ispyb_utils import (
    get_current_time_string,
    get_ispyb_config,
)
from mx_bluesky.common.utils.exceptions import ISPyBDepositionNotMade

RobotActionID = int


class BearerAuth(AuthBase):
    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers["authorization"] = "Bearer " + self.token
        return r


def _get_base_url_and_token() -> tuple[str, str]:
    config = configparser.ConfigParser()
    conf = get_ispyb_config()
    config.read(conf)
    expeye_config = config["expeye"]
    return expeye_config["url"], expeye_config["token"]


def _send_and_get_response(auth, url, data, send_func) -> dict:
    response = send_func(url, auth=auth, json=data)
    if not response.ok:
        try:
            resp_txt = str(response.json())
        except JSONDecodeError:
            resp_txt = str(response)
        raise ISPyBDepositionNotMade(f"Could not write {data} to {url}: {resp_txt}")
    return response.json()


@dataclass
class BLSample:
    container_id: int
    bl_sample_id: int
    bl_sample_status: str | None


class BLSampleStatus(StrEnum):
    # The sample has been loaded
    LOADED = "LOADED"
    # Problem with the sample e.g. pin too long/short
    ERROR_SAMPLE = "ERROR - sample"
    # Any other general error
    ERROR_BEAMLINE = "ERROR - beamline"


assert all(len(value) <= 20 for value in BLSampleStatus), (
    "Column size limit of 20 for BLSampleStatus"
)


class ExpeyeInteraction:
    """Exposes functionality from the Expeye core API"""

    CREATE_ROBOT_ACTION = "/proposals/{proposal}/sessions/{visit_number}/robot-actions"
    UPDATE_ROBOT_ACTION = "/robot-actions/{action_id}"

    def __init__(self) -> None:
        url, token = _get_base_url_and_token()
        self._base_url = url
        self._auth = BearerAuth(token)

    def start_load(
        self,
        proposal_reference: str,
        visit_number: int,
        sample_id: int,
        dewar_location: int,
        container_location: int,
    ) -> RobotActionID:
        """Create a robot load entry in ispyb.

        Args:
            proposal_reference (str): The proposal of the experiment e.g. cm37235
            visit_number (int): The visit number for the proposal, usually this can be
                                found added to the end of the proposal e.g. the data for
                                visit number 2 of proposal cm37235 is in cm37235-2
            sample_id (int): The id of the sample in the database
            dewar_location (int): Which puck in the dewar the sample is in
            container_location (int): Which pin in that puck has the sample

        Returns:
            RobotActionID: The id of the robot load action that is created
        """
        url = self._base_url + self.CREATE_ROBOT_ACTION.format(
            proposal=proposal_reference, visit_number=visit_number
        )

        data = {
            "startTimestamp": get_current_time_string(),
            "sampleId": sample_id,
            "actionType": "LOAD",
            "containerLocation": container_location,
            "dewarLocation": dewar_location,
        }
        response = _send_and_get_response(self._auth, url, data, post)
        return response["robotActionId"]

    def update_barcode_and_snapshots(
        self,
        action_id: RobotActionID,
        barcode: str,
        snapshot_before_path: str,
        snapshot_after_path: str,
    ):
        """Update the barcode and snapshots of an existing robot action.

        Args:
            action_id (RobotActionID): The id of the action to update
            barcode (str): The barcode to give the action
            snapshot_before_path (str): Path to the snapshot before robot load
            snapshot_after_path (str): Path to the snapshot after robot load
        """
        url = self._base_url + self.UPDATE_ROBOT_ACTION.format(action_id=action_id)

        data = {
            "sampleBarcode": barcode,
            "xtalSnapshotBefore": snapshot_before_path,
            "xtalSnapshotAfter": snapshot_after_path,
        }
        _send_and_get_response(self._auth, url, data, patch)

    def end_load(self, action_id: RobotActionID, status: str, reason: str):
        """Finish an existing robot action, providing final information about how it went

        Args:
            action_id (RobotActionID): The action to finish.
            status (str): The status of the action at the end, "success" for success,
                          otherwise error
            reason (str): If the status is in error than the reason for that error
        """
        url = self._base_url + self.UPDATE_ROBOT_ACTION.format(action_id=action_id)

        run_status = "SUCCESS" if status == "success" else "ERROR"

        data = {
            "endTimestamp": get_current_time_string(),
            "status": run_status,
            "message": reason[:255] if reason else "",
        }
        _send_and_get_response(self._auth, url, data, patch)

    def update_sample_status(
        self, bl_sample_id: int, bl_sample_status: BLSampleStatus
    ) -> BLSample:
        """Update the blSampleStatus of a sample.
        Args:
            bl_sample_id: The sample ID
            bl_sample_status: The sample status
            status_message: An optional message
        Returns:
             The updated sample
        """
        data = {"blSampleStatus": (str(bl_sample_status))}
        response = _send_and_get_response(
            self._auth, self._base_url + f"/samples/{bl_sample_id}", data, patch
        )
        return self._sample_from_json(response)

    def _sample_from_json(self, response) -> BLSample:
        return BLSample(
            bl_sample_id=response["blSampleId"],
            bl_sample_status=response["blSampleStatus"],
            container_id=response["containerId"],
        )



from mx_bluesky.beamlines.i24.serial.web_gui_plans.general_plans import (
    gui_gonio_move_on_click,
    gui_move_detector,
    gui_set_parameters,
    gui_sleep,
    gui_stage_move_on_click,
)

from .extruder.i24ssx_Extruder_Collect_py3v2 import (
    enter_hutch,
    initialise_extruder,
    laser_check,
    run_extruder_plan,
)
from .fixed_target.i24ssx_Chip_Collect_py3v1 import run_fixed_target_plan
from .fixed_target.i24ssx_Chip_Manager_py3v1 import (
    block_check,
    cs_maker,
    cs_reset,
    define_current_chip,
    fiducial,
    initialise_stages,
    laser_control,
    load_lite_map,
    load_stock_map,
    moveto,
    moveto_preset,
    pumpprobe_calc,
)
from .log import clean_up_log_config_at_end, setup_collection_logs
from .setup_beamline.setup_detector import setup_detector_stage

__all__ = [
    "setup_detector_stage",
    "run_extruder_plan",
    "initialise_extruder",
    "enter_hutch",
    "laser_check",
    "run_fixed_target_plan",
    "moveto",
    "moveto_preset",
    "block_check",
    "cs_maker",
    "cs_reset",
    "define_current_chip",
    "fiducial",
    "initialise_stages",
    "laser_control",
    "load_lite_map",
    "load_stock_map",
    "pumpprobe_calc",
    "setup_collection_logs",
    "clean_up_log_config_at_end",
    # GUI plans
    "gui_stage_move_on_click",
    "gui_gonio_move_on_click",
    "gui_sleep",
    "gui_move_detector",
    "gui_set_parameters",
]



from collections.abc import Callable
from functools import wraps

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
from bluesky.utils import MsgGenerator
from dodal.devices.zebra.zebra import (
    ArmDemand,
    EncEnum,
    I03Axes,
    RotationDirection,
    Zebra,
)
from dodal.devices.zebra.zebra_controlled_shutter import (
    ZebraShutter,
    ZebraShutterControl,
)

from mx_bluesky.common.utils.log import LOGGER

ZEBRA_STATUS_TIMEOUT = 30


def bluesky_retry(func: Callable):
    """Decorator that will retry the decorated plan if it fails.

    Use this with care as it knows nothing about the state of the world when things fail.
    If it is possible that your plan fails when the beamline is in a transient state that
    the plan could not act on do not use this decorator without doing some more intelligent
    clean up.

    You should avoid using this decorator often in general production as it hides errors,
    instead it should be used only for debugging these underlying errors.
    """

    @wraps(func)
    def newfunc(*args, **kwargs):
        def log_and_retry(exception):
            LOGGER.error(f"Function {func.__name__} failed with {exception}, retrying")
            yield from func(*args, **kwargs)

        yield from bpp.contingency_wrapper(
            func(*args, **kwargs), except_plan=log_and_retry, auto_raise=False
        )

    return newfunc


def arm_zebra(zebra: Zebra):
    yield from bps.abs_set(zebra.pc.arm, ArmDemand.ARM, wait=True)


def tidy_up_zebra_after_rotation_scan(
    zebra: Zebra,
    zebra_shutter: ZebraShutter,
    group="tidy_up_zebra_after_rotation",
    wait=True,
):
    yield from bps.abs_set(zebra.pc.arm, ArmDemand.DISARM, group=group)
    yield from bps.abs_set(
        zebra_shutter.control_mode, ZebraShutterControl.MANUAL, group=group
    )
    if wait:
        yield from bps.wait(group, timeout=ZEBRA_STATUS_TIMEOUT)


def set_shutter_auto_input(zebra: Zebra, input: int, group="set_shutter_trigger"):
    """Set the signal that controls the shutter. We use the second input to the
    Zebra's AND2 gate for this input. ZebraShutter control mode must be in auto for this input to take control

    For more details see the ZebraShutter device."""
    auto_gate = zebra.mapping.AND_GATE_FOR_AUTO_SHUTTER
    auto_shutter_control = zebra.logic_gates.and_gates[auto_gate]
    yield from bps.abs_set(auto_shutter_control.sources[2], input, group)


def configure_zebra_and_shutter_for_auto_shutter(
    zebra: Zebra, zebra_shutter: ZebraShutter, input: int, group="use_automatic_shutter"
):
    """Set the shutter to auto mode, and configure the zebra to trigger the shutter on
    an input source. For the input, use one of the source constants in zebra.py

    When the shutter is in auto/manual, logic in EPICS sets the Zebra's
    SOFT_IN1 to low/high respectively. The Zebra's AND2 gate should be used to control the shutter while in auto mode.
    To do this, we need (AND2 = SOFT_IN1 AND input), where input is the zebra signal we want to control the shutter when in auto mode.
    """
    # See https://github.com/DiamondLightSource/dodal/issues/813 for better typing here.

    # Set shutter to auto mode
    yield from bps.abs_set(
        zebra_shutter.control_mode, ZebraShutterControl.AUTO, group=group
    )

    auto_gate = zebra.mapping.AND_GATE_FOR_AUTO_SHUTTER

    # Set first input of AND2 gate to SOFT_IN1, which is high when shutter is in auto mode
    # Note the Zebra should ALWAYS be setup this way. See https://github.com/DiamondLightSource/mx-bluesky/issues/551
    yield from bps.abs_set(
        zebra.logic_gates.and_gates[auto_gate].sources[1],
        zebra.mapping.sources.SOFT_IN1,
        group=group,
    )

    # Set the second input of AND2 gate to the requested zebra input source
    yield from set_shutter_auto_input(zebra, input, group=group)


@bluesky_retry
def setup_zebra_for_rotation(
    zebra: Zebra,
    zebra_shutter: ZebraShutter,
    axis: EncEnum = I03Axes.OMEGA,
    start_angle: float = 0,
    scan_width: float = 360,
    shutter_opening_deg: float = 2.5,
    shutter_opening_s: float = 0.04,
    direction: RotationDirection = RotationDirection.POSITIVE,
    group: str = "setup_zebra_for_rotation",
    wait: bool = True,
):
    """Set up the Zebra to collect a rotation dataset. Any plan using this is
    responsible for setting the smargon velocity appropriately so that the desired
    image width is achieved with the exposure time given here.

    Parameters:
        zebra:              The zebra device to use
        axis:               I03 axes enum representing which axis to use for position
                            compare. Currently always omega.
        start_angle:        Position at which the scan should begin, in degrees.
        scan_width:         Total angle through which to collect, in degrees.
        shutter_opening_deg:How many degrees of rotation it takes for the fast shutter
                            to open. Increases the gate width.
        shutter_opening_s:  How many seconds it takes for the fast shutter to open. The
                            detector pulse is delayed after the shutter signal by this
                            amount.
        direction:          RotationDirection enum for positive or negative.
                            Defaults to Positive.
        group:              A name for the group of statuses generated
        wait:               Block until all the settings have completed
    """

    if not isinstance(direction, RotationDirection):
        raise ValueError(
            "Disallowed rotation direction provided to Zebra setup plan. "
            "Use RotationDirection.POSITIVE or RotationDirection.NEGATIVE."
        )
    yield from bps.abs_set(zebra.pc.dir, direction.value, group=group)
    LOGGER.info("ZEBRA SETUP: START")
    # Set gate start, adjust for shutter opening time if necessary
    LOGGER.info(f"ZEBRA SETUP: degrees to adjust for shutter = {shutter_opening_deg}")
    LOGGER.info(f"ZEBRA SETUP: start angle start: {start_angle}")
    LOGGER.info(f"ZEBRA SETUP: start angle adjusted, gate start set to: {start_angle}")
    yield from bps.abs_set(zebra.pc.gate_start, start_angle, group=group)
    # set gate width to total width
    yield from bps.abs_set(
        zebra.pc.gate_width, scan_width + shutter_opening_deg, group=group
    )
    LOGGER.info(
        f"Pulse start set to shutter open time, set to: {abs(shutter_opening_s)}"
    )
    yield from bps.abs_set(zebra.pc.pulse_start, abs(shutter_opening_s), group=group)
    # Set gate position to be angle of interest
    yield from bps.abs_set(zebra.pc.gate_trigger, axis.value, group=group)
    # Set shutter to automatic and to trigger via PC_GATE
    yield from configure_zebra_and_shutter_for_auto_shutter(
        zebra, zebra_shutter, zebra.mapping.sources.PC_GATE, group=group
    )
    # Trigger the detector with a pulse
    yield from bps.abs_set(
        zebra.output.out_pvs[zebra.mapping.outputs.TTL_DETECTOR],
        zebra.mapping.sources.PC_PULSE,
        group=group,
    )
    # Don't use the fluorescence detector
    yield from bps.abs_set(
        zebra.output.out_pvs[zebra.mapping.outputs.TTL_XSPRESS3],
        zebra.mapping.sources.DISCONNECT,
        group=group,
    )
    yield from bps.abs_set(
        zebra.output.pulse_1.input, zebra.mapping.sources.DISCONNECT, group=group
    )
    LOGGER.info(f"ZEBRA SETUP: END - {'' if wait else 'not'} waiting for completion")
    if wait:
        yield from bps.wait(group, timeout=ZEBRA_STATUS_TIMEOUT)


@bluesky_retry
def setup_zebra_for_gridscan(
    zebra: Zebra,
    zebra_shutter: ZebraShutter,
    group="setup_zebra_for_gridscan",
    wait=True,
):
    # Set shutter to automatic and to trigger via motion controller GPIO signal (IN4_TTL)
    yield from configure_zebra_and_shutter_for_auto_shutter(
        zebra, zebra_shutter, zebra.mapping.sources.IN4_TTL, group=group
    )

    yield from bps.abs_set(
        zebra.output.out_pvs[zebra.mapping.outputs.TTL_DETECTOR],
        zebra.mapping.sources.IN3_TTL,
        group=group,
    )
    yield from bps.abs_set(
        zebra.output.out_pvs[zebra.mapping.outputs.TTL_XSPRESS3],
        zebra.mapping.sources.DISCONNECT,
        group=group,
    )
    yield from bps.abs_set(
        zebra.output.pulse_1.input, zebra.mapping.sources.DISCONNECT, group=group
    )

    if wait:
        yield from bps.wait(group, timeout=ZEBRA_STATUS_TIMEOUT)


@bluesky_retry
def tidy_up_zebra_after_gridscan(
    zebra: Zebra,
    zebra_shutter: ZebraShutter,
    group="tidy_up_zebra_after_gridscan",
    wait=True,
) -> MsgGenerator:
    yield from bps.abs_set(
        zebra.output.out_pvs[zebra.mapping.outputs.TTL_DETECTOR],
        zebra.mapping.sources.PC_PULSE,
        group=group,
    )
    yield from bps.abs_set(
        zebra_shutter.control_mode, ZebraShutterControl.MANUAL, group=group
    )
    yield from set_shutter_auto_input(zebra, zebra.mapping.sources.PC_GATE, group=group)

    if wait:
        yield from bps.wait(group, timeout=ZEBRA_STATUS_TIMEOUT)


@bluesky_retry
def setup_zebra_for_panda_flyscan(
    zebra: Zebra,
    zebra_shutter: ZebraShutter,
    group="setup_zebra_for_panda_flyscan",
    wait=True,
):
    # Forwards eiger trigger signal from panda
    yield from bps.abs_set(
        zebra.output.out_pvs[zebra.mapping.outputs.TTL_DETECTOR],
        zebra.mapping.sources.IN1_TTL,
        group=group,
    )

    # Set shutter to automatic and to trigger via motion controller GPIO signal (IN4_TTL)
    yield from configure_zebra_and_shutter_for_auto_shutter(
        zebra, zebra_shutter, zebra.mapping.sources.IN4_TTL, group=group
    )

    yield from bps.abs_set(
        zebra.output.out_pvs[zebra.mapping.outputs.TTL_XSPRESS3],
        zebra.mapping.sources.DISCONNECT,
        group=group,
    )

    yield from bps.abs_set(
        zebra.output.out_pvs[zebra.mapping.outputs.TTL_PANDA],
        zebra.mapping.sources.IN3_TTL,
        group=group,
    )  # Tells panda that motion is beginning/changing direction

    if wait:
        yield from bps.wait(group, timeout=ZEBRA_STATUS_TIMEOUT)



from __future__ import annotations

from itertools import dropwhile, takewhile
from typing import Any
from unittest.mock import ANY, MagicMock, Mock, call, patch

import pytest
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from dodal.devices.aperturescatterguard import ApertureScatterguard, ApertureValue
from dodal.devices.backlight import BacklightPosition
from dodal.devices.detector.detector_motion import ShutterState
from dodal.devices.i03.beamstop import BeamstopPositions
from dodal.devices.oav.oav_parameters import OAVParameters
from dodal.devices.smargon import Smargon
from dodal.devices.synchrotron import SynchrotronMode
from dodal.devices.xbpm_feedback import Pause
from dodal.devices.zebra.zebra import RotationDirection, Zebra
from dodal.devices.zebra.zebra_controlled_shutter import ZebraShutterControl
from ophyd_async.testing import get_mock_put, set_mock_value

from mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback import (
    ZocaloCallback,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_store import IspybIds
from mx_bluesky.common.parameters.constants import DocDescriptorNames
from mx_bluesky.hyperion.device_setup_plans.check_beamstop import BeamstopException
from mx_bluesky.hyperion.experiment_plans.oav_snapshot_plan import (
    OAV_SNAPSHOT_GROUP,
)
from mx_bluesky.hyperion.experiment_plans.rotation_scan_plan import (
    RotationMotionProfile,
    RotationScanComposite,
    calculate_motion_profile,
    rotation_scan,
    rotation_scan_plan,
)
from mx_bluesky.hyperion.external_interaction.callbacks.rotation.ispyb_callback import (
    RotationISPyBCallback,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.rotation import RotationScan

from .conftest import fake_read

TEST_OFFSET = 1
TEST_SHUTTER_OPENING_DEGREES = 2.5


def do_rotation_main_plan_for_tests(
    run_eng: RunEngine,
    expt_params: RotationScan,
    devices: RotationScanComposite,
    motion_values: RotationMotionProfile,
):
    with patch(
        "bluesky.preprocessors.__read_and_stash_a_motor",
        fake_read,
    ):
        run_eng(
            rotation_scan_plan(devices, expt_params, motion_values),
        )


@pytest.fixture
def run_full_rotation_plan(
    RE: RunEngine,
    test_rotation_params: RotationScan,
    fake_create_rotation_devices: RotationScanComposite,
    oav_parameters_for_rotation: OAVParameters,
) -> RotationScanComposite:
    with patch(
        "bluesky.preprocessors.__read_and_stash_a_motor",
        fake_read,
    ):
        RE(
            rotation_scan(
                fake_create_rotation_devices,
                test_rotation_params,
                oav_parameters_for_rotation,
            ),
        )
        return fake_create_rotation_devices


@pytest.fixture
def motion_values(test_rotation_params: RotationScan):
    return calculate_motion_profile(
        test_rotation_params,
        0.005,  # time for acceleration
        222,
    )


def setup_and_run_rotation_plan_for_tests(
    RE: RunEngine,
    test_params: RotationScan,
    fake_create_rotation_devices: RotationScanComposite,
    motion_values,
):
    with patch("bluesky.plan_stubs.wait", autospec=True):
        do_rotation_main_plan_for_tests(
            RE, test_params, fake_create_rotation_devices, motion_values
        )

    return {
        "RE_with_subs": RE,
        "test_rotation_params": test_params,
        "smargon": fake_create_rotation_devices.smargon,
        "zebra": fake_create_rotation_devices.zebra,
    }


@pytest.fixture
def setup_and_run_rotation_plan_for_tests_standard(
    RE: RunEngine,
    test_rotation_params: RotationScan,
    fake_create_rotation_devices: RotationScanComposite,
    motion_values: RotationMotionProfile,
):
    return setup_and_run_rotation_plan_for_tests(
        RE, test_rotation_params, fake_create_rotation_devices, motion_values
    )


@pytest.fixture
def setup_and_run_rotation_plan_for_tests_nomove(
    RE: RunEngine,
    test_rotation_params_nomove: RotationScan,
    fake_create_rotation_devices: RotationScanComposite,
    motion_values: RotationMotionProfile,
):
    return setup_and_run_rotation_plan_for_tests(
        RE, test_rotation_params_nomove, fake_create_rotation_devices, motion_values
    )


def test_rotation_scan_calculations(test_rotation_params: RotationScan):
    test_rotation_params.features.omega_flip = False
    test_rotation_params.exposure_time_s = 0.2
    test_rotation_params.omega_start_deg = 10

    motion_values = calculate_motion_profile(
        test_rotation_params,
        0.005,  # time for acceleration
        224,
    )

    assert motion_values.direction == "Negative"
    assert motion_values.start_scan_deg == 10

    assert motion_values.speed_for_rotation_deg_s == 0.5  # 0.1 deg per 0.2 sec
    assert motion_values.shutter_time_s == 0.6
    assert motion_values.shutter_opening_deg == 0.3  # distance moved in 0.6 s

    # 1.5 * distance moved in time for accel (fudge)
    assert motion_values.acceleration_offset_deg == 0.00375
    assert motion_values.start_motion_deg == 10.00375

    assert motion_values.total_exposure_s == 360
    assert motion_values.scan_width_deg == 180
    assert motion_values.distance_to_move_deg == -180.3075


@patch(
    "dodal.common.beamlines.beamline_utils.active_device_is_same_type",
    lambda a, b: True,
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.rotation_scan_plan.rotation_scan_plan",
    autospec=True,
)
def test_rotation_scan(
    plan: MagicMock,
    RE: RunEngine,
    test_rotation_params: RotationScan,
    fake_create_rotation_devices: RotationScanComposite,
    oav_parameters_for_rotation: OAVParameters,
):
    composite = fake_create_rotation_devices
    RE(rotation_scan(composite, test_rotation_params, oav_parameters_for_rotation))
    composite.eiger.do_arm.set.assert_called()  # type: ignore
    composite.eiger.unstage.assert_called()  # type: ignore


def test_rotation_plan_runs(
    setup_and_run_rotation_plan_for_tests_standard: dict[str, Any],
) -> None:
    RE: RunEngine = setup_and_run_rotation_plan_for_tests_standard["RE_with_subs"]
    assert RE._exit_status == "success"


async def test_rotation_plan_zebra_settings(
    setup_and_run_rotation_plan_for_tests_standard: dict[str, Any],
) -> None:
    zebra: Zebra = setup_and_run_rotation_plan_for_tests_standard["zebra"]
    params: RotationScan = setup_and_run_rotation_plan_for_tests_standard[
        "test_rotation_params"
    ]

    assert await zebra.pc.gate_start.get_value() == params.omega_start_deg
    assert await zebra.pc.pulse_start.get_value() == params.shutter_opening_time_s


async def test_full_rotation_plan_smargon_settings(
    run_full_rotation_plan: RotationScanComposite,
    test_rotation_params: RotationScan,
) -> None:
    smargon: Smargon = run_full_rotation_plan.smargon
    params: RotationScan = test_rotation_params

    test_max_velocity = await smargon.omega.max_velocity.get_value()

    omega_set: MagicMock = get_mock_put(smargon.omega.user_setpoint)
    omega_velocity_set: MagicMock = get_mock_put(smargon.omega.velocity)
    rotation_speed = params.rotation_increment_deg / params.exposure_time_s

    assert await smargon.phi.user_setpoint.get_value() == params.phi_start_deg
    assert await smargon.chi.user_setpoint.get_value() == params.chi_start_deg
    assert await smargon.x.user_setpoint.get_value() == params.x_start_um / 1000  # type: ignore
    assert await smargon.y.user_setpoint.get_value() == params.y_start_um / 1000  # type: ignore
    assert await smargon.z.user_setpoint.get_value() == params.z_start_um / 1000  # type: ignore
    assert (
        # 4 * snapshots, restore omega, 1 * rotation sweep
        omega_set.call_count == 4 + 1 + 1
    )
    # 1 to max vel in outer plan, 1 to max vel in setup_oav_snapshot_plan, 1 set before rotation, 1 restore in cleanup plan
    assert omega_velocity_set.call_count == 4
    assert omega_velocity_set.call_args_list == [
        call(test_max_velocity, wait=True),
        call(test_max_velocity, wait=True),
        call(rotation_speed, wait=True),
        call(test_max_velocity, wait=True),
    ]


async def test_rotation_plan_moves_aperture_correctly(
    run_full_rotation_plan: RotationScanComposite,
) -> None:
    aperture_scatterguard: ApertureScatterguard = (
        run_full_rotation_plan.aperture_scatterguard
    )
    assert (
        await aperture_scatterguard.selected_aperture.get_value() == ApertureValue.SMALL
    )


async def test_rotation_plan_smargon_doesnt_move_xyz_if_not_given_in_params(
    setup_and_run_rotation_plan_for_tests_nomove: dict[str, Any],
) -> None:
    smargon: Smargon = setup_and_run_rotation_plan_for_tests_nomove["smargon"]
    params: RotationScan = setup_and_run_rotation_plan_for_tests_nomove[
        "test_rotation_params"
    ]
    assert params.phi_start_deg is None
    assert params.chi_start_deg is None
    assert params.x_start_um is None
    assert params.y_start_um is None
    assert params.z_start_um is None
    for motor in [smargon.phi, smargon.chi, smargon.x, smargon.y, smargon.z]:
        get_mock_put(motor.user_setpoint).assert_not_called()  # type: ignore


@patch(
    "mx_bluesky.hyperion.experiment_plans.rotation_scan_plan._cleanup_plan",
    autospec=True,
)
@patch("bluesky.plan_stubs.wait", autospec=True)
def test_cleanup_happens(
    bps_wait: MagicMock,
    cleanup_plan: MagicMock,
    RE: RunEngine,
    test_rotation_params: RotationScan,
    fake_create_rotation_devices: RotationScanComposite,
    motion_values: RotationMotionProfile,
    oav_parameters_for_rotation: OAVParameters,
):
    class MyTestException(Exception):
        pass

    failing_set = MagicMock(
        side_effect=MyTestException("Experiment fails because this is a test")
    )

    with patch.object(fake_create_rotation_devices.smargon.omega, "set", failing_set):
        # check main subplan part fails
        with pytest.raises(MyTestException):
            RE(
                rotation_scan_plan(
                    fake_create_rotation_devices, test_rotation_params, motion_values
                )
            )
        cleanup_plan.assert_not_called()
        # check that failure is handled in composite plan
        with pytest.raises(MyTestException) as exc:
            RE(
                rotation_scan(
                    fake_create_rotation_devices,
                    test_rotation_params,
                    oav_parameters_for_rotation,
                )
            )
        assert "Experiment fails because this is a test" in exc.value.args[0]
        cleanup_plan.assert_called_once()


def test_rotation_plan_reads_hardware(
    fake_create_rotation_devices: RotationScanComposite,
    test_rotation_params,
    motion_values,
    sim_run_engine_for_rotation: RunEngineSimulator,
):
    _add_sim_handlers_for_normal_operation(
        fake_create_rotation_devices, sim_run_engine_for_rotation
    )
    msgs = sim_run_engine_for_rotation.simulate_plan(
        rotation_scan_plan(
            fake_create_rotation_devices, test_rotation_params, motion_values
        )
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "create"
        and msg.kwargs["name"] == CONST.DESCRIPTORS.HARDWARE_READ_PRE,
    )
    msgs_in_event = list(takewhile(lambda msg: msg.command != "save", msgs))
    assert_message_and_return_remaining(
        msgs_in_event, lambda msg: msg.command == "read" and msg.obj.name == "smargon"
    )


@pytest.fixture
def rotation_scan_simulated_messages(
    sim_run_engine: RunEngineSimulator,
    fake_create_rotation_devices: RotationScanComposite,
    test_rotation_params: RotationScan,
    oav_parameters_for_rotation: OAVParameters,
):
    _add_sim_handlers_for_normal_operation(fake_create_rotation_devices, sim_run_engine)

    return sim_run_engine.simulate_plan(
        rotation_scan(
            fake_create_rotation_devices,
            test_rotation_params,
            oav_parameters_for_rotation,
        )
    )


def test_rotation_scan_initialises_detector_distance_shutter_and_tx_fraction(
    rotation_scan_simulated_messages,
    test_rotation_params: RotationScan,
):
    msgs = assert_message_and_return_remaining(
        rotation_scan_simulated_messages,
        lambda msg: msg.command == "set"
        and msg.args[0] == test_rotation_params.detector_distance_mm
        and msg.obj.name == "detector_motion-z"
        and msg.kwargs["group"] == CONST.WAIT.ROTATION_READY_FOR_DC,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.args[0] == ShutterState.OPEN
        and msg.obj.name == "detector_motion-shutter"
        and msg.kwargs["group"] == CONST.WAIT.ROTATION_READY_FOR_DC,
    )
    assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "wait"
        and msg.kwargs["group"] == CONST.WAIT.ROTATION_READY_FOR_DC,
    )


def test_rotation_scan_triggers_xbpm_then_pauses_xbpm_and_sets_transmission(
    rotation_scan_simulated_messages,
    test_rotation_params: RotationScan,
):
    msgs = assert_message_and_return_remaining(
        rotation_scan_simulated_messages,
        lambda msg: msg.command == "trigger" and msg.obj.name == "xbpm_feedback",
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "xbpm_feedback-pause_feedback"
        and msg.args[0] == Pause.PAUSE.value,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "attenuator"
        and msg.args[0] == test_rotation_params.transmission_frac,
    )


def test_rotation_scan_does_not_change_transmission_back_until_after_data_collected(
    rotation_scan_simulated_messages,
    test_rotation_params: RotationScan,
):
    msgs = assert_message_and_return_remaining(
        rotation_scan_simulated_messages,
        lambda msg: msg.command == "unstage" and msg.obj.name == "eiger",
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "xbpm_feedback-pause_feedback"
        and msg.args[0] == Pause.RUN.value,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "attenuator"
        and msg.args[0] == 1.0,
    )


def test_rotation_scan_moves_gonio_to_start_before_snapshots(
    rotation_scan_simulated_messages,
):
    msgs = assert_message_and_return_remaining(
        rotation_scan_simulated_messages,
        lambda msg: msg.command == "wait"
        and msg.kwargs["group"] == CONST.WAIT.MOVE_GONIO_TO_START,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "wait"
        and msg.kwargs["group"] == CONST.WAIT.READY_FOR_OAV,
    )


def test_rotation_scan_moves_aperture_in_backlight_out_after_snapshots_before_rotation(
    rotation_scan_simulated_messages,
):
    msgs = assert_message_and_return_remaining(
        rotation_scan_simulated_messages,
        lambda msg: msg.command == "create"
        and msg.kwargs["name"] == DocDescriptorNames.OAV_ROTATION_SNAPSHOT_TRIGGERED,
    )
    msgs = assert_message_and_return_remaining(msgs, lambda msg: msg.command == "save")
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "aperture_scatterguard"
        and msg.args[0] == ApertureValue.SMALL
        and msg.kwargs["group"] == CONST.WAIT.ROTATION_READY_FOR_DC,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "backlight"
        and msg.args[0] == BacklightPosition.OUT
        and msg.kwargs["group"] == CONST.WAIT.ROTATION_READY_FOR_DC,
    )
    assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "wait"
        and msg.kwargs["group"] == CONST.WAIT.ROTATION_READY_FOR_DC,
    )


def test_rotation_scan_resets_omega_waits_for_sample_env_complete_after_snapshots_before_hw_read(
    test_rotation_params: RotationScan, rotation_scan_simulated_messages
):
    msgs = assert_message_and_return_remaining(
        rotation_scan_simulated_messages,
        lambda msg: msg.command == "create"
        and msg.kwargs["name"] == DocDescriptorNames.OAV_ROTATION_SNAPSHOT_TRIGGERED,
    )
    msgs = assert_message_and_return_remaining(msgs, lambda msg: msg.command == "save")
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "smargon-omega"
        and msg.args[0] == test_rotation_params.omega_start_deg
        and msg.kwargs["group"] == CONST.WAIT.ROTATION_READY_FOR_DC,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "wait"
        and msg.kwargs["group"] == CONST.WAIT.ROTATION_READY_FOR_DC,
    )
    assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "create"
        and msg.kwargs["name"] == CONST.DESCRIPTORS.ZOCALO_HW_READ,
    )


def test_rotation_snapshot_setup_called_to_move_backlight_in_aperture_out_before_triggering(
    rotation_scan_simulated_messages,
):
    msgs = assert_message_and_return_remaining(
        rotation_scan_simulated_messages,
        lambda msg: msg.command == "set"
        and msg.obj.name == "backlight"
        and msg.args[0] == BacklightPosition.IN
        and msg.kwargs["group"] == CONST.WAIT.READY_FOR_OAV,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "aperture_scatterguard"
        and msg.args[0] == ApertureValue.OUT_OF_BEAM
        and msg.kwargs["group"] == CONST.WAIT.READY_FOR_OAV,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "wait"
        and msg.kwargs["group"] == CONST.WAIT.READY_FOR_OAV,
    )
    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "trigger" and msg.obj.name == "oav-snapshot"
    )


def test_rotation_scan_skips_init_backlight_aperture_and_snapshots_if_snapshot_params_specified(
    fake_create_rotation_devices: RotationScanComposite,
    sim_run_engine: RunEngineSimulator,
    test_rotation_params: RotationScan,
    oav_parameters_for_rotation: OAVParameters,
):
    _add_sim_handlers_for_normal_operation(fake_create_rotation_devices, sim_run_engine)
    test_rotation_params.snapshot_omegas_deg = None

    msgs = sim_run_engine.simulate_plan(
        rotation_scan(
            fake_create_rotation_devices,
            test_rotation_params,
            oav_parameters_for_rotation,
        )
    )
    assert not [
        msg for msg in msgs if msg.kwargs.get("group", None) == CONST.WAIT.READY_FOR_OAV
    ]
    assert not [
        msg for msg in msgs if msg.kwargs.get("group", None) == OAV_SNAPSHOT_GROUP
    ]
    assert (
        len(
            [
                msg
                for msg in msgs
                if msg.command == "set"
                and msg.obj.name == "smargon-omega"
                and msg.kwargs["group"] == CONST.WAIT.ROTATION_READY_FOR_DC
            ]
        )
        == 1
    )


def _add_sim_handlers_for_normal_operation(
    fake_create_rotation_devices, sim_run_engine: RunEngineSimulator
):
    sim_run_engine.add_handler(
        "read",
        lambda msg: {"values": {"value": SynchrotronMode.USER}},
        "synchrotron-synchrotron_mode",
    )
    sim_run_engine.add_handler(
        "read",
        lambda msg: {"values": {"value": -1}},
        "synchrotron-top_up_start_countdown",
    )
    sim_run_engine.add_handler(
        "read", lambda msg: {"smargon-omega": {"value": -1}}, "smargon-omega"
    )


def test_rotation_scan_turns_shutter_to_auto_with_pc_gate_then_back_to_manual(
    fake_create_rotation_devices: RotationScanComposite,
    sim_run_engine: RunEngineSimulator,
    test_rotation_params: RotationScan,
    oav_parameters_for_rotation: OAVParameters,
):
    _add_sim_handlers_for_normal_operation(fake_create_rotation_devices, sim_run_engine)
    msgs = sim_run_engine.simulate_plan(
        rotation_scan(
            fake_create_rotation_devices,
            test_rotation_params,
            oav_parameters_for_rotation,
        )
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "sample_shutter-control_mode"
        and msg.args[0] == ZebraShutterControl.AUTO,  # type:ignore
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "zebra-logic_gates-and_gates-2-sources-1"
        and msg.args[0] == fake_create_rotation_devices.zebra.mapping.sources.SOFT_IN1,  # type:ignore
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "zebra-logic_gates-and_gates-2-sources-2"
        and msg.args[0] == fake_create_rotation_devices.zebra.mapping.sources.PC_GATE,  # type:ignore
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "sample_shutter-control_mode"
        and msg.args[0] == ZebraShutterControl.MANUAL,  # type:ignore
    )


def test_rotation_scan_arms_detector_and_takes_snapshots_whilst_arming(
    rotation_scan_simulated_messages,
    test_rotation_params,
    fake_create_rotation_devices,
    oav_parameters_for_rotation,
):
    composite = fake_create_rotation_devices
    msgs = assert_message_and_return_remaining(
        rotation_scan_simulated_messages,
        lambda msg: (
            msg.command == "open_run"
            and "BeamDrawingCallback" in msg.kwargs.get("activate_callbacks", [])
        ),
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "eiger_do_arm"
        and msg.args[0] == 1
        and msg.kwargs["group"] == CONST.WAIT.ROTATION_READY_FOR_DC,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj is composite.oav.snapshot.directory
        and msg.args[0] == str(test_rotation_params.snapshot_directory),
    )
    for omega in test_rotation_params.snapshot_omegas_deg:
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "set"
            and msg.obj is composite.smargon.omega
            and msg.args[0] == omega,
        )
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "set"
            and msg.obj is composite.oav.snapshot.filename
            and f"_oav_snapshot_{omega:.0f}" in msg.args[0],
        )
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "trigger" and msg.obj.name == "oav-snapshot",
        )
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "create"
            and msg.kwargs["name"]
            == DocDescriptorNames.OAV_ROTATION_SNAPSHOT_TRIGGERED,
        )
        msgs = assert_message_and_return_remaining(
            msgs, lambda msg: msg.command == "read" and msg.obj is composite.oav
        )
        msgs = assert_message_and_return_remaining(
            msgs, lambda msg: msg.command == "save"
        )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "wait"
        and msg.kwargs["group"] == CONST.WAIT.ROTATION_READY_FOR_DC,
    )


@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.rotation.ispyb_callback.StoreInIspyb"
)
def test_rotation_scan_correctly_triggers_ispyb_callback(
    mock_store_in_ispyb,
    RE: RunEngine,
    test_rotation_params: RotationScan,
    fake_create_rotation_devices: RotationScanComposite,
    oav_parameters_for_rotation: OAVParameters,
):
    mock_ispyb_callback = RotationISPyBCallback()
    RE.subscribe(mock_ispyb_callback)
    with (
        patch("bluesky.plan_stubs.wait", autospec=True),
        patch(
            "bluesky.preprocessors.__read_and_stash_a_motor",
            fake_read,
        ),
    ):
        RE(
            rotation_scan(
                fake_create_rotation_devices,
                test_rotation_params,
                oav_parameters_for_rotation,
            ),
        )
    mock_store_in_ispyb.assert_called()


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback.ZocaloTrigger"
)
@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.rotation.ispyb_callback.StoreInIspyb"
)
def test_rotation_scan_correctly_triggers_zocalo_callback(
    mock_store_in_ispyb,
    mock_zocalo_interactor,
    RE: RunEngine,
    test_rotation_params: RotationScan,
    fake_create_rotation_devices: RotationScanComposite,
    oav_parameters_for_rotation: OAVParameters,
):
    mock_zocalo_callback = ZocaloCallback(CONST.PLAN.ROTATION_MAIN, "env")
    mock_ispyb_callback = RotationISPyBCallback(emit=mock_zocalo_callback)
    mock_store_in_ispyb.return_value.update_deposition.return_value = IspybIds(
        data_collection_ids=(0, 1)
    )
    RE.subscribe(mock_ispyb_callback)
    with (
        patch("bluesky.plan_stubs.wait", autospec=True),
        patch(
            "bluesky.preprocessors.__read_and_stash_a_motor",
            fake_read,
        ),
    ):
        RE(
            rotation_scan(
                fake_create_rotation_devices,
                test_rotation_params,
                oav_parameters_for_rotation,
            ),
        )
    mock_zocalo_interactor.return_value.run_start.assert_called_once()


def test_rotation_scan_fails_with_exception_when_no_beamstop(
    sim_run_engine: RunEngineSimulator,
    fake_create_rotation_devices: RotationScanComposite,
    test_rotation_params: RotationScan,
    oav_parameters_for_rotation: OAVParameters,
):
    sim_run_engine.add_read_handler_for(
        fake_create_rotation_devices.beamstop.selected_pos, BeamstopPositions.UNKNOWN
    )
    with pytest.raises(BeamstopException):
        sim_run_engine.simulate_plan(
            rotation_scan(
                fake_create_rotation_devices,
                test_rotation_params,
                oav_parameters_for_rotation,
            )
        )


@pytest.mark.parametrize(
    "omega_flip, rotation_direction, expected_start_angle, "
    "expected_start_angle_with_runup, expected_zebra_direction",
    [
        # see https://github.com/DiamondLightSource/mx-bluesky/issues/247
        # GDA behaviour is such that positive angles in the request result in
        # negative motor angles, but positive angles in the resulting nexus file
        # Should replicate GDA Output exactly
        [True, RotationDirection.POSITIVE, -30, -29.85, RotationDirection.NEGATIVE],
        # Should replicate GDA Output, except with /entry/data/transformation/omega
        # +1, 0, 0 instead of -1, 0, 0
        [False, RotationDirection.NEGATIVE, 30, 30.15, RotationDirection.NEGATIVE],
        [True, RotationDirection.NEGATIVE, -30, -30.15, RotationDirection.POSITIVE],
        [False, RotationDirection.POSITIVE, 30, 29.85, RotationDirection.POSITIVE],
    ],
)
@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback.ZocaloTrigger",
    MagicMock(),
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.rotation_scan_plan.setup_zebra_for_rotation"
)
def test_rotation_scan_plan_with_omega_flip_inverts_motor_movements_but_not_event_params(
    mock_setup_zebra_for_rotation: MagicMock,
    omega_flip: bool,
    rotation_direction: RotationDirection,
    expected_start_angle: float,
    expected_start_angle_with_runup: float,
    expected_zebra_direction: RotationDirection,
    test_rotation_params: RotationScan,
    fake_create_rotation_devices: RotationScanComposite,
    oav_parameters_for_rotation: OAVParameters,
    RE: RunEngine,
):
    test_rotation_params.features.omega_flip = omega_flip
    test_rotation_params.rotation_direction = rotation_direction
    test_rotation_params.omega_start_deg = 30
    mock_callback = Mock(spec=RotationISPyBCallback)
    RE.subscribe(mock_callback)
    omega_put = get_mock_put(fake_create_rotation_devices.smargon.omega.user_setpoint)
    set_mock_value(fake_create_rotation_devices.smargon.omega.acceleration_time, 0.1)
    with (
        patch("bluesky.plan_stubs.wait", autospec=True),
        patch(
            "bluesky.preprocessors.__read_and_stash_a_motor",
            fake_read,
        ),
    ):
        RE(
            rotation_scan(
                fake_create_rotation_devices,
                test_rotation_params,
                oav_parameters_for_rotation,
            ),
        )

    assert omega_put.mock_calls[0:5] == [
        call(0, wait=True),
        call(90, wait=True),
        call(180, wait=True),
        call(270, wait=True),
        call(expected_start_angle_with_runup, wait=True),
    ]
    mock_setup_zebra_for_rotation.assert_called_once_with(
        fake_create_rotation_devices.zebra,
        fake_create_rotation_devices.sample_shutter,
        start_angle=expected_start_angle,
        scan_width=180,
        direction=expected_zebra_direction,
        shutter_opening_deg=ANY,
        shutter_opening_s=ANY,
        group="setup_zebra",
    )
    rotation_outer_start_event = next(
        dropwhile(
            lambda _: _.args[0] != "start"
            or _.args[1].get("subplan_name") != CONST.PLAN.ROTATION_OUTER,
            mock_callback.mock_calls,
        )
    )
    event_params = RotationScan.model_validate_json(
        rotation_outer_start_event.args[1]["mx_bluesky_parameters"]
    )
    # event params are not transformed
    assert event_params.omega_start_deg == 30
    assert event_params.rotation_direction == rotation_direction
    assert event_params.rotation_increment_deg == 0.1
    assert event_params.scan_width_deg == 180
    assert event_params.features.omega_flip == omega_flip


def test_rotation_scan_does_not_verify_undulator_gap_until_before_run(
    rotation_scan_simulated_messages,
    test_rotation_params: RotationScan,
):
    msgs = rotation_scan_simulated_messages
    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "set" and msg.obj.name == "undulator"
    )
    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "open_run"
    )



from enum import StrEnum

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp

from mx_bluesky.common.external_interaction.callbacks.sample_handling.sample_handling_callback import (
    SampleHandlingCallback,
)
from mx_bluesky.common.utils.exceptions import SampleException


class SampleStatusExceptionType(StrEnum):
    BEAMLINE = "Beamline"
    SAMPLE = "Sample"


@bpp.subs_decorator(SampleHandlingCallback())
def deposit_sample_error(exception_type: SampleStatusExceptionType, sample_id: int):
    @bpp.run_decorator(
        md={
            "metadata": {"sample_id": sample_id},
            "activate_callbacks": ["SampleHandlingCallback"],
        }
    )
    def _inner():
        yield from bps.null()
        if exception_type == SampleStatusExceptionType.BEAMLINE:
            raise AssertionError()
        elif exception_type == SampleStatusExceptionType.SAMPLE:
            raise SampleException

    yield from _inner()


@bpp.subs_decorator(SampleHandlingCallback(record_loaded_on_success=True))
def deposit_loaded_sample(sample_id: int):
    @bpp.run_decorator(
        md={
            "metadata": {"sample_id": sample_id},
            "activate_callbacks": ["SampleHandlingCallback"],
        }
    )
    def _inner():
        yield from bps.null()

    yield from _inner()



from __future__ import annotations

from abc import abstractmethod
from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING, Any, TypeVar, cast

from dodal.beamline_specific_utils.i03 import beam_size_from_aperture
from dodal.devices.detector import DetectorParams
from dodal.devices.detector.det_resolution import resolution
from dodal.devices.synchrotron import SynchrotronMode

from mx_bluesky.common.external_interaction.callbacks.common.logging_callback import (
    format_doc_for_log,
)
from mx_bluesky.common.external_interaction.callbacks.common.plan_reactive_callback import (
    PlanReactiveCallback,
)
from mx_bluesky.common.external_interaction.ispyb.data_model import (
    DataCollectionInfo,
    DataCollectionPositionInfo,
    ScanDataInfo,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_store import (
    IspybIds,
    StoreInIspyb,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_utils import get_ispyb_config
from mx_bluesky.common.parameters.components import DiffractionExperimentWithSample
from mx_bluesky.common.parameters.constants import DocDescriptorNames, SimConstants
from mx_bluesky.common.utils.log import ISPYB_ZOCALO_CALLBACK_LOGGER, set_dcgid_tag
from mx_bluesky.common.utils.utils import convert_eV_to_angstrom

D = TypeVar("D")
if TYPE_CHECKING:
    from event_model.documents import Event, EventDescriptor, RunStart, RunStop


def _update_based_on_energy(
    doc: Event,
    detector_params: DetectorParams,
    data_collection_info: DataCollectionInfo,
):
    """If energy has been read as part of this reading then add it into the data
    collection info along with the other fields that depend on it."""
    if energy_kev := doc["data"].get("dcm-energy_in_kev", None):
        energy_ev = energy_kev * 1000
        wavelength_angstroms = convert_eV_to_angstrom(energy_ev)
        data_collection_info.wavelength = wavelength_angstroms
        data_collection_info.resolution = resolution(
            detector_params,
            wavelength_angstroms,
            detector_params.detector_distance,
        )
    return data_collection_info


class BaseISPyBCallback(PlanReactiveCallback):
    def __init__(
        self,
        *,
        emit: Callable[..., Any] | None = None,
    ) -> None:
        """Subclasses should run super().__init__() with parameters, then set
        self.ispyb to the type of ispyb relevant to the experiment and define the type
        for self.ispyb_ids."""
        ISPYB_ZOCALO_CALLBACK_LOGGER.debug("Initialising ISPyB callback")
        super().__init__(log=ISPYB_ZOCALO_CALLBACK_LOGGER, emit=emit)
        self._oav_snapshot_event_idx: int = 0
        self.params: DiffractionExperimentWithSample | None = None
        self.ispyb: StoreInIspyb
        self.descriptors: dict[str, EventDescriptor] = {}
        self.ispyb_config = get_ispyb_config()
        if (
            self.ispyb_config == SimConstants.ISPYB_CONFIG
            or self.ispyb_config == SimConstants.DEV_ISPYB_DATABASE_CFG
        ):
            ISPYB_ZOCALO_CALLBACK_LOGGER.warning(
                f"{self.__class__} using dev ISPyB config: {self.ispyb_config}. If you"
                "want to use the real database, please set the ISPYB_CONFIG_PATH "
                "environment variable."
            )
        self.uid_to_finalize_on: str | None = None
        self.ispyb_ids: IspybIds = IspybIds()
        self.log = ISPYB_ZOCALO_CALLBACK_LOGGER

    def activity_gated_start(self, doc: RunStart):
        self._oav_snapshot_event_idx = 0
        return self._tag_doc(doc)

    def activity_gated_descriptor(self, doc: EventDescriptor):
        self.descriptors[doc["uid"]] = doc
        return self._tag_doc(doc)

    def activity_gated_event(self, doc: Event) -> Event:
        """Subclasses should extend this to add a call to set_dcig_tag from
        hyperion.log"""
        ISPYB_ZOCALO_CALLBACK_LOGGER.debug("ISPyB handler received event document.")
        assert self.ispyb is not None, "ISPyB deposition wasn't initialised!"
        assert self.params is not None, "ISPyB handler didn't receive parameters!"

        event_descriptor = self.descriptors.get(doc["descriptor"])
        if event_descriptor is None:
            ISPYB_ZOCALO_CALLBACK_LOGGER.warning(
                f"Ispyb handler {self} received event doc {format_doc_for_log(doc)} and "
                "has no corresponding descriptor record"
            )
            return doc
        match event_descriptor.get("name"):
            case DocDescriptorNames.HARDWARE_READ_PRE:
                scan_data_infos = self._handle_ispyb_hardware_read(doc)
            case DocDescriptorNames.HARDWARE_READ_DURING:
                scan_data_infos = self._handle_ispyb_transmission_flux_read(doc)
            case _:
                return self._tag_doc(doc)
        self.ispyb_ids = self.ispyb.update_deposition(self.ispyb_ids, scan_data_infos)
        ISPYB_ZOCALO_CALLBACK_LOGGER.info(f"Received ISPYB IDs: {self.ispyb_ids}")
        return self._tag_doc(doc)

    def _handle_ispyb_hardware_read(self, doc) -> Sequence[ScanDataInfo]:
        assert self.params, "Event handled before activity_gated_start received params"
        ISPYB_ZOCALO_CALLBACK_LOGGER.info(
            "ISPyB handler received event from read hardware"
        )
        synchrotron_mode = doc["data"]["synchrotron-synchrotron_mode"]
        assert isinstance(synchrotron_mode, SynchrotronMode)

        hwscan_data_collection_info = DataCollectionInfo(
            undulator_gap1=doc["data"]["undulator-current_gap"],
            synchrotron_mode=synchrotron_mode.value,
            slitgap_horizontal=doc["data"]["s4_slit_gaps-xgap"],
            slitgap_vertical=doc["data"]["s4_slit_gaps-ygap"],
        )
        hwscan_data_collection_info = _update_based_on_energy(
            doc, self.params.detector_params, hwscan_data_collection_info
        )
        hwscan_position_info = DataCollectionPositionInfo(
            pos_x=float(doc["data"]["smargon-x"]),
            pos_y=float(doc["data"]["smargon-y"]),
            pos_z=float(doc["data"]["smargon-z"]),
        )
        scan_data_infos = self.populate_info_for_update(
            hwscan_data_collection_info, hwscan_position_info, self.params
        )
        ISPYB_ZOCALO_CALLBACK_LOGGER.info(
            "Updating ispyb data collection after hardware read."
        )
        return scan_data_infos

    def _handle_ispyb_transmission_flux_read(self, doc) -> Sequence[ScanDataInfo]:
        assert self.params
        aperture = doc["data"]["aperture_scatterguard-selected_aperture"]
        aperture_radius = doc["data"]["aperture_scatterguard-radius"]
        beamsize = beam_size_from_aperture(aperture_radius)
        beamsize_x_mm = beamsize.x_um / 1000 if beamsize.x_um else None
        beamsize_y_mm = beamsize.y_um / 1000 if beamsize.y_um else None
        hwscan_data_collection_info = DataCollectionInfo(
            beamsize_at_samplex=beamsize_x_mm,
            beamsize_at_sampley=beamsize_y_mm,
            focal_spot_size_at_samplex=beamsize_x_mm,
            focal_spot_size_at_sampley=beamsize_y_mm,
            flux=doc["data"]["flux-flux_reading"],
        )
        if transmission := doc["data"]["attenuator-actual_transmission"]:
            # Ispyb wants the transmission in a percentage, we use fractions
            hwscan_data_collection_info.transmission = transmission * 100
        hwscan_data_collection_info = _update_based_on_energy(
            doc, self.params.detector_params, hwscan_data_collection_info
        )
        scan_data_infos = self.populate_info_for_update(
            hwscan_data_collection_info, None, self.params
        )
        ISPYB_ZOCALO_CALLBACK_LOGGER.info(
            "Updating ispyb data collection after flux read."
        )
        self.append_to_comment(f"Aperture: {aperture}. ")
        return scan_data_infos

    @abstractmethod
    def populate_info_for_update(
        self,
        event_sourced_data_collection_info: DataCollectionInfo,
        event_sourced_position_info: DataCollectionPositionInfo | None,
        params: DiffractionExperimentWithSample,
    ) -> Sequence[ScanDataInfo]:
        pass

    def activity_gated_stop(self, doc: RunStop) -> RunStop:
        """Subclasses must check that they are recieving a stop document for the correct
        uid to use this method!"""
        assert self.ispyb is not None, (
            "ISPyB handler received stop document, but deposition object doesn't exist!"
        )
        ISPYB_ZOCALO_CALLBACK_LOGGER.debug("ISPyB handler received stop document.")
        exit_status = (
            doc.get("exit_status") or "Exit status not available in stop document!"
        )
        reason = doc.get("reason") or ""
        set_dcgid_tag(None)
        try:
            self.ispyb.end_deposition(self.ispyb_ids, exit_status, reason)
        except Exception as e:
            ISPYB_ZOCALO_CALLBACK_LOGGER.warning(
                f"Failed to finalise ISPyB deposition on stop document: {format_doc_for_log(doc)} with exception: {e}"
            )
        return self._tag_doc(doc)

    def _append_to_comment(self, id: int, comment: str) -> None:
        assert self.ispyb is not None
        try:
            self.ispyb.append_to_comment(id, comment)
        except TypeError:
            ISPYB_ZOCALO_CALLBACK_LOGGER.warning(
                "ISPyB deposition not initialised, can't update comment."
            )

    def append_to_comment(self, comment: str):
        for id in self.ispyb_ids.data_collection_ids:
            self._append_to_comment(id, comment)

    def _tag_doc(self, doc: D) -> D:
        assert isinstance(doc, dict)
        if self.ispyb_ids:
            doc["ispyb_dcids"] = self.ispyb_ids.data_collection_ids
        return cast(D, doc)
